﻿// COPYRIGHT 2014, 2015 by the Open Rails project.
// 
// This file is part of Open Rails.
// 
// Open Rails is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Open Rails is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Open Rails.  If not, see <http://www.gnu.org/licenses/>.
//
"use strict";

var PathCharts = function () {

    var my = {};

    // Public configuration properties. Can be modified externally
    // The name of the HTML element that contains the header. If it is not defined, then no canvas drawing will occur
    my.canvasName = "";
    // The name of the HTML element that contains the header. If it is not defined, then no heading will be changed will occur
    my.headerName = "";
    // The amount of pixels per vertical km. This will be used to change the height of the canvas 
    my.pixelsPerKm = 50;
    // The drawing can either use km/m or miles/feet
    my.useImperialUnits = false;
    // Drawing the curvature chart can either use straight lines of rounded curves.
    my.useRoundedCurvatureCharts = true;
    // The parsed data from the json file. This allows access to the object in case you need it.
    my.pathChartData = false;

    // Various subcharts are drawn on the canvas. Each will take the full height, but only a part of the width.
    // That width is defined by defining the relative position of the left side and of the right side.
    my.curvatureSubChartRelativeLeft = 0.05;
    my.curvatureSubChartRelativeRight = 0.15;
    my.speedlimitsSubChartRelativeLeft = 0.15;
    my.speedlimitsSubChartRelativeRight = 0.25;
    my.distanceSubChartRelativeLeft = 0.25;
    my.distanceSubChartRelativeRight = 0.35;
    my.heightSubChartRelativeLeft = 0.35;
    my.heightSubChartRelativeRight = 0.9;
    my.milemarkersSubChartRelativeLeft = 0.9;
    my.milemarkersSubChartRelativeRight = 0.99;

    // Public method(s)
    my.Draw = function (pathChartData, callback) {
        my.pathChartData = pathChartData;
        _changeHeader();
        _drawCharts();
    }

    // private properties & methods
    var TYPE_STATION = 1;
    var TYPE_MILEPOST = 2;
    var TYPE_SPEEDLIMIT_FORWARD = 3;
    var TYPE_SPEEDLIMIT_REVERSE = 4;

    var _changeHeader = function () {
        var headerNode = document.getElementById(my.headerName);
        if (!headerNode) {
            return;
        }

        while (headerNode.firstChild) {
            headerNode.removeChild(headerNode.firstChild);
        }
        headerNode.appendChild(document.createTextNode(my.pathChartData.PathName));
    }

    var _drawCharts = function () {
        var canvas = document.getElementById(my.canvasName);
        if (!canvas) {
            return;
        }
        if (!canvas.getContext) {
            return;
        }

        var charts = {};
        charts["curvature"] = new CurvatureChart(my.pathChartData, my.curvatureSubChartRelativeLeft, my.curvatureSubChartRelativeRight);
        charts["distance"] = new DistanceChart(my.pathChartData, my.distanceSubChartRelativeLeft, my.distanceSubChartRelativeRight);
        charts["height"] = new HeightChart(my.pathChartData, my.heightSubChartRelativeLeft, my.heightSubChartRelativeRight);
        charts["milemarkers"] = new MilemarkerChart(my.pathChartData, my.milemarkersSubChartRelativeLeft, my.milemarkersSubChartRelativeRight);
        charts["speedlimits"] = new SpeedlimitsChart(my.pathChartData, my.speedlimitsSubChartRelativeLeft, my.speedlimitsSubChartRelativeRight);

        var canvasHeight = my.pathChartData.PointWithMaxima.DistanceAlongPath * my.pixelsPerKm / 1000;
        canvas.height = canvasHeight;
        for (var chart in charts) {
            charts[chart].Draw(canvas);
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  Charts baseclass       /////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    function SubChart() {
    }

    SubChart.prototype.SetConstructorArguments = function(pathData, relativeL, relativeR) {
        this.pathData = pathData;
        this.relativeL = relativeL;
        this.relativeR = relativeR;

        this.minX = this.pathData.PointWithMinima.DistanceAlongPath;
        this.maxX = this.pathData.PointWithMaxima.DistanceAlongPath;
        this.minY = this.getYvalue(this.pathData.PointWithMinima);
        this.maxY = this.getYvalue(this.pathData.PointWithMaxima);
        if (this.minY == this.maxY) {
            this.minY -= 10;
            this.maxY += 10;
        }
    }

    SubChart.prototype.Draw = function (canvas) {
        this.canvasWidth = canvas.width;
        this.canvasHeight = canvas.height;
        this.ctx = canvas.getContext("2d");

        this.DrawSubChart();
    }

    SubChart.prototype.DrawDataPolyLine = function () {
        var pathPoints = this.pathData.PathChartPoints.$values;
        this.ctx.beginPath();
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            var sourceX = chartPoint.DistanceAlongPath;
            var sourceY = this.getYvalue(chartPoint);
            //alert(this.sourceYtoPixelsX(sourceY));
            this.ctx.lineTo(this.sourceYtoPixelsX(sourceY), this.sourceXtoPixelsY(sourceX));
        }
        this.ctx.stroke();

    }

    SubChart.prototype.DrawLine = function (x1, y1, x2, y2, color) {
        this.ctx.save();
        this.ctx.strokeStyle = color;
        this.ctx.beginPath();
        this.ctx.moveTo(x1, y1);
        this.ctx.lineTo(x2, y2);
        this.ctx.stroke();
        this.ctx.restore();
    }

    SubChart.prototype.sourceXtoPixelsY = function (sourceX) {
        // x in source data, so along the path. 
        var topMargin = 10;
        var bottomMargin = 10;
        var effectiveHeight = this.canvasHeight - (topMargin + bottomMargin);
        return bottomMargin + (sourceX - this.minX) / (this.maxX - this.minX) * effectiveHeight;
    }

    SubChart.prototype.sourceYtoPixelsX = function (sourceY) {
        // y in source data, so the values of interest. top and bottom are also in source y-direction
        var leftX = 10 + this.relativeL * this.canvasWidth;
        var rightX = 10 + this.relativeR * this.canvasWidth;
        var effectiveWidth = rightX - leftX;
        return leftX + (sourceY - this.minY) / (this.maxY - this.minY) * effectiveWidth;
    }

    SubChart.prototype.getYvalue = function () {
        // just a default implementation
        return 0;
    }

    SubChart.prototype.toString = function () {
        return "sub";
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  Height subchart        /////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    HeightChart.prototype = new SubChart();
    HeightChart.prototype.constructor = SubChart;
    function HeightChart(pathData, relativeL, relativeR) {
        SubChart.prototype.SetConstructorArguments.apply(this, arguments);
    }

    HeightChart.prototype.DrawSubChart = function () {
        this.CleanScaling();
        this.DrawGridLines();
        this.PrintGrades();
        this.DrawDataPolyLine(false);
        this.PrintStationNames();
    }

    HeightChart.prototype.getYvalue = function (pathChartPoint) {
        return pathChartPoint.HeightM;
    }
        
    HeightChart.prototype.CleanScaling = function () {
        if (my.useImperialUnits)
        {
            this.niceScale = new NiceScaling(this.minY, this.maxY, 0.3048, "ft");
        }
        else
        {
            this.niceScale = new NiceScaling(this.minY, this.maxY, 1.0, "m");
        }
        this.minY = this.niceScale.ValueMin;
        this.maxY = this.niceScale.ValueMax;
    }

    HeightChart.prototype.DrawGridLines = function () {
        this.ctx.save();
        this.ctx.textBaseline = "alphabetic";
        this.ctx.textAlign = "center";
        for (var i=0; i < this.niceScale.NiceValues.length; i++)
        {
            var niceValue = this.niceScale.NiceValues[i];
            this.DrawLine(
                this.sourceYtoPixelsX(niceValue), this.sourceXtoPixelsY(this.minX),
                this.sourceYtoPixelsX(niceValue), this.sourceXtoPixelsY(this.maxX), '#808080');

            this.ctx.fillText(niceValue.toFixed(0) + this.niceScale.Unit, this.sourceYtoPixelsX(niceValue), this.sourceXtoPixelsY(this.minX))
        }
        this.ctx.restore();
    }

    HeightChart.prototype.PrintStationNames = function () {
        this.ctx.save();
        this.ctx.textBaseline = "middle";
        this.ctx.textAlign = "left";
        var lastDrawnText = "";
        var pathPoints = this.pathData.PathChartPoints.$values;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            if (chartPoint.TrackItemType == TYPE_STATION) {
                var sourceX = chartPoint.DistanceAlongPath;
                var sourceY = this.getYvalue(chartPoint);
                if (chartPoint.TrackItemText == lastDrawnText) {
                    lastDrawnText = "";
                }
                else {
                    lastDrawnText = chartPoint.TrackItemText;
                    this.ctx.fillText(lastDrawnText, this.sourceYtoPixelsX(sourceY), this.sourceXtoPixelsY(sourceX))
                }
            }
        }
        this.ctx.restore();
    }

    HeightChart.prototype.PrintGrades = function () {
        this.ctx.save();
        this.ctx.textBaseline = "middle";
        this.ctx.textAlign = "right";
        var textoffset = this.ctx.measureText("-0.0").width;

        var lastY = this.sourceXtoPixelsY(0);
        var zero = 0;
        var lastTextToDraw = zero.toFixed(1);

        var gradeChangeIndexes = [0];

        var pathPoints = this.pathData.PathChartPoints.$values;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];

            var newY = this.sourceXtoPixelsY(chartPoint.DistanceAlongPath);
            var newTextToDraw = chartPoint.GradePercent.toFixed(1);

            if (lastTextToDraw != newTextToDraw) {
                this.ctx.fillText(lastTextToDraw, this.sourceYtoPixelsX(this.minY)+textoffset, (lastY + newY) / 2);
                lastY = newY;
                lastTextToDraw = newTextToDraw;
                gradeChangeIndexes.push(i);
            }
        }
        this.ctx.restore();

        // and now we draw lines from base to height chart where the grade changed
        this.ctx.save();
        this.ctx.strokeStyle = '#808080';
        this.ctx.beginPath();
        for (var i = 0; i < gradeChangeIndexes.length; i++) {
            var chartPoint = pathPoints[gradeChangeIndexes[i]];
            this.ctx.moveTo(this.sourceYtoPixelsX(this.minY), this.sourceXtoPixelsY(chartPoint.DistanceAlongPath));
            this.ctx.lineTo(this.sourceYtoPixelsX(chartPoint.HeightM), this.sourceXtoPixelsY(chartPoint.DistanceAlongPath));
        }
        this.ctx.stroke();
        this.ctx.restore();
    }

    HeightChart.prototype.toString = function () {
        return "height";
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  Curvature subchart              ////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    CurvatureChart.prototype = new SubChart();
    CurvatureChart.prototype.constructor = SubChart;
    function CurvatureChart(pathData, relativeL, relativeR) {
        SubChart.prototype.SetConstructorArguments.apply(this, arguments);
    }

    CurvatureChart.prototype.DrawSubChart = function () {
        this.DrawZeroGrid();
        if (my.useRoundedCurvatureCharts) {
            this.DrawCurvatureRounded();
        }
        else {
            this.DrawCurvatureSimple();
        }
        this.PrintDegreesTurned();
    }

    CurvatureChart.prototype.DrawZeroGrid = function () {
        this.DrawLine(
            this.sourceYtoPixelsX(0), this.sourceXtoPixelsY(this.minX),
            this.sourceYtoPixelsX(0), this.sourceXtoPixelsY(this.maxX), '#808080');
    }

    CurvatureChart.prototype.DrawCurvatureSimple = function () {
        var pathPoints = this.pathData.PathChartPoints.$values;
        this.ctx.beginPath();
        var lastX = this.sourceYtoPixelsX(sourceY);
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            var sourceX = chartPoint.DistanceAlongPath;
            var sourceY = this.getYvalue(chartPoint);
            var newX = this.sourceYtoPixelsX(sourceY);
            if (newX != lastX) {
                this.ctx.lineTo(lastX, this.sourceXtoPixelsY(sourceX));
            }
            this.ctx.lineTo(newX, this.sourceXtoPixelsY(sourceX));
            lastX = newX;
        }
        this.ctx.stroke();
    }

    CurvatureChart.prototype.DrawCurvatureRounded = function () {
        var pathPoints = this.pathData.PathChartPoints.$values;
        this.ctx.beginPath();
        var lastX = this.sourceYtoPixelsX(sourceY);
        var lastSourceX = 0;
        var centerX = this.sourceYtoPixelsX(0);
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            var sourceX = chartPoint.DistanceAlongPath;
            var sourceY = this.getYvalue(chartPoint);
            var newX = this.sourceYtoPixelsX(sourceY);
            if (newX != lastX || i == pathPoints.length - 1) { // a bit dirty, but we do need the last point also
                var startY = this.sourceXtoPixelsY(lastSourceX);
                var stopY = this.sourceXtoPixelsY(sourceX);
                var deltaY = stopY - startY;
                var deltaX = Math.abs(lastX-centerX);
                if (deltaY < 2*deltaX) {
                    // There is not really enough room. So make a small arc.
                    //this.ctx.arc(centerX, (startY + stopY) / 2, deltaY / 2, -Math.PI / 2, Math.PI / 2, lastX < centerX);
                    // Use quadratic bezier curves
                    this.ctx.quadraticCurveTo(lastX, startY, lastX, (startY + stopY) / 2);
                    this.ctx.quadraticCurveTo(lastX, stopY,  centerX, stopY);
                }
                else {
                    this.ctx.quadraticCurveTo(lastX, startY, lastX, startY + deltaX);
                    this.ctx.lineTo(lastX, stopY - deltaX);
                    this.ctx.quadraticCurveTo(lastX, stopY, centerX, stopY);
                }
                lastX = newX;
                lastSourceX = sourceX;
            }
        }
        this.ctx.stroke();
    }

    CurvatureChart.prototype.PrintDegreesTurned = function () {
        this.ctx.save();
        this.ctx.textBaseline = "middle";
        this.ctx.textAlign = "right";
        //Offsets are at the right side of the text
        var degreeSymbol = "\xB0";
        var offsetLeft = this.sourceYtoPixelsX(0) - 3;
        var offsetRight = this.sourceYtoPixelsX(0) + 3 + this.ctx.measureText("-10" + degreeSymbol).width;

        var pathPoints = this.pathData.PathChartPoints.$values;
        var degreesTurned = 0;
        var curveSignOld = 0;
        var lastSourceX = 0;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            var curveSignNew = Math.sign(chartPoint.Curvature);
            if (curveSignNew != curveSignOld) {
                if (curveSignOld != 0) {
                    var textOffset = (curveSignOld > 0) ? offsetLeft : offsetRight;
                    this.ctx.fillText(degreesTurned.toFixed(0) + degreeSymbol, textOffset, this.sourceXtoPixelsY((lastSourceX + chartPoint.DistanceAlongPath)/2));
                }
                degreesTurned = 0;
                curveSignOld = curveSignNew;
                lastSourceX = chartPoint.DistanceAlongPath;
            }
            degreesTurned += chartPoint.DistanceAlongNextSection * chartPoint.Curvature * 180 / Math.PI;
        }
        this.ctx.restore();
    }

    CurvatureChart.prototype.getYvalue = function (pathChartPoint) {
        //return pathChartPoint.Curvature;
        return 10*Math.sign(pathChartPoint.Curvature);
    }
   
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  Distance along path subchart    ////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    DistanceChart.prototype = new SubChart();
    DistanceChart.prototype.construSubChartctor = SubChart;
    function DistanceChart(pathData, relativeL, relativeR) {
        SubChart.prototype.SetConstructorArguments.apply(this, arguments);
    }

    DistanceChart.prototype.DrawSubChart = function () {
        this.ctx.save();
        this.ctx.textBaseline = "middle";
        this.ctx.textAlign = "center";
        var distanceUnitInMeters, unit;
        if (my.useImperialUnits) {
            distanceUnitInMeters = 1609.34;
            unit = "mi";
        }
        else {
            distanceUnitInMeters = 1000;
            unit = "km";
        }
        var distanceMax = this.maxX / distanceUnitInMeters;
        for (var distance = 0; distance <= distanceMax; distance++) {
            this.ctx.fillText(distance.toFixed(0)+unit, this.sourceYtoPixelsX(0), this.sourceXtoPixelsY(distance*distanceUnitInMeters))

        }
        this.ctx.restore();
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  MileMarker subchart    /////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    MilemarkerChart.prototype = new SubChart();
    MilemarkerChart.prototype.constructor = SubChart;
    function MilemarkerChart(pathData, relativeL, relativeR) {
        SubChart.prototype.SetConstructorArguments.apply(this, arguments);
    }

    MilemarkerChart.prototype.DrawSubChart = function () {
        this.DrawMileMarkers();
        this.DrawCirclesAroundMarkers();
    }

    MilemarkerChart.prototype.DrawMileMarkers = function () {
        this.ctx.save();
        this.ctx.textBaseline = "middle";
        this.ctx.textAlign = "center";
        var pathPoints = this.pathData.PathChartPoints.$values;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            if (chartPoint.TrackItemType == TYPE_MILEPOST) {
                var sourceX = chartPoint.DistanceAlongPath;
                var sourceY = this.getYvalue(chartPoint);
                this.ctx.fillText(chartPoint.TrackItemText, this.sourceYtoPixelsX(sourceY), this.sourceXtoPixelsY(sourceX))
            }
        }
        this.ctx.restore();
    }

    MilemarkerChart.prototype.DrawCirclesAroundMarkers = function () {
        this.ctx.beginPath();
        var radius = 15;
        var pathPoints = this.pathData.PathChartPoints.$values;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            if (chartPoint.TrackItemType == TYPE_MILEPOST) {
                var sourceX = chartPoint.DistanceAlongPath;
                var sourceY = this.getYvalue(chartPoint);
                var pixelsX = this.sourceYtoPixelsX(sourceY);
                var pixelsY = this.sourceXtoPixelsY(sourceX);
                this.ctx.moveTo(pixelsX + radius, pixelsY);
                this.ctx.arc(pixelsX, pixelsY, radius, 0, 2*Math.PI);
            }
        }
        this.ctx.stroke();

    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  SpeedLimits subchart    /////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    SpeedlimitsChart.prototype = new SubChart();
    SpeedlimitsChart.prototype.constructor = SubChart;
    function SpeedlimitsChart(pathData, relativeL, relativeR) {
        SubChart.prototype.SetConstructorArguments.apply(this, arguments);
    }

    SpeedlimitsChart.prototype.DrawSubChart = function () {
        this.DrawSpeedMarkers();
        this.DrawDirectionAroundMarkers();
    }

    SpeedlimitsChart.prototype.DrawSpeedMarkers = function () {
        this.ctx.save();
        this.ctx.textBaseline = "middle";
        this.ctx.textAlign = "center";
        var pathPoints = this.pathData.PathChartPoints.$values;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            if (chartPoint.TrackItemType == TYPE_SPEEDLIMIT_FORWARD || chartPoint.TrackItemType == TYPE_SPEEDLIMIT_REVERSE) {
                var sourceX = chartPoint.DistanceAlongPath;
                var sourceY = this.getYvalue(chartPoint);
                this.ctx.fillText(chartPoint.TrackItemText, this.sourceYtoPixelsX(sourceY), this.sourceXtoPixelsY(sourceX))
            }
        }
        this.ctx.restore();
    }

    SpeedlimitsChart.prototype.DrawDirectionAroundMarkers = function () {
        var markerSize = this.ctx.measureText("100").width;
        this.ctx.beginPath();
        var pathPoints = this.pathData.PathChartPoints.$values;
        for (var i = 0; i < pathPoints.length; i++) {
            var chartPoint = pathPoints[i];
            if (chartPoint.TrackItemType == TYPE_SPEEDLIMIT_FORWARD || chartPoint.TrackItemType == TYPE_SPEEDLIMIT_REVERSE) {
                var delta = (chartPoint.TrackItemType == TYPE_SPEEDLIMIT_FORWARD ? 0.5 : -0.5) * markerSize;

                var sourceX = chartPoint.DistanceAlongPath;
                var sourceY = this.getYvalue(chartPoint);
                var pixelsX = this.sourceYtoPixelsX(sourceY);
                var pixelsY = this.sourceXtoPixelsY(sourceX);
                this.ctx.moveTo(pixelsX - delta, pixelsY + delta);
                this.ctx.lineTo(pixelsX - delta, pixelsY - delta);
                this.ctx.lineTo(pixelsX + delta, pixelsY - delta);
                this.ctx.lineTo(pixelsX + delta, pixelsY + delta);
                this.ctx.lineTo(pixelsX        , pixelsY + 2*delta);
                this.ctx.lineTo(pixelsX - delta, pixelsY + delta);
            }
        }
        this.ctx.stroke();

    }

    SpeedlimitsChart.prototype.getYvalue = function (pathChartPoint) {
        if (pathChartPoint.TrackItemType == TYPE_SPEEDLIMIT_FORWARD) {
            return 10;
        }
        if (pathChartPoint.TrackItemType == TYPE_SPEEDLIMIT_REVERSE) {
            return -10;
        }
        return 0;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////  Nice scaling           /////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    function NiceScaling(valueMin, valueMax, scale, unit, minNumberOfTicks, strict) {
        if (!minNumberOfTicks) {
            minNumberOfTicks = 4;
        }
        if (valueMin == valueMax)
        {
            // exceptional situation, there is no range. So make one up.
            if (valueMin == 0)
            {
                valueMin = -1;
                valueMax = +1;
            }
            else
            {
                if (valueMin < 0)
                {
                    valueMax = 0;
                }
                else
                {
                    valueMin = 0;
                }
            }
        }


        this.Unit = unit;
        var valueScaledMin = valueMin/scale;
        var valueScaledMax = valueMax/scale;
        var subDivisionMax = (valueScaledMax - valueScaledMin)/minNumberOfTicks;
        var sub10log = Math.floor(Math.log10(subDivisionMax));

        var power10 = Math.pow(10.0, sub10log);
        var subDivisionMaxNice = subDivisionMax / power10;

        //Now we move to decimal to make sure we do not have rounding errors
        this.Scale = scale;
        var valueScaledStep = (power10) * (
            (subDivisionMaxNice >= 5) ? 5 :
            (subDivisionMaxNice >= 4) ? 4 :
            (subDivisionMaxNice >= 2) ? 2 :
            1);
        var valueScaledStart = valueScaledStep * (Math.floor((valueScaledMin) / valueScaledStep));
        if (strict && (valueScaledStart < valueScaledMin))
        {
            valueScaledStart = valueScaledStep * (1+Math.floor((valueScaledMin) / valueScaledStep));
        }
            
        var i = 0;
        this.NiceValues = new Array();
        var valueScaled = valueScaledStart;
        while (valueScaled < valueScaledMax) {
            this.NiceValues.push(valueScaled * this.Scale);
            i++;
            valueScaled = valueScaledStart + i*valueScaledStep;
        }
        this.NiceValues.push(valueScaled * this.Scale);

        this.ValueMin = valueScaledStart * this.Scale;
        this.ValueMax = valueScaled      * this.Scale;
    }

    Math.log10 = Math.log10 || function (x) {
        return Math.log(x) / Math.LN10;
    };
    Math.sign = Math.sign || function (x) {
        if (+x === x) { // check if a number was given
            return (x === 0) ? x : (x > 0) ? 1 : -1;
        }
        return NaN;
    }

    return my;
}();